@@ -58,6 +58,14 @@ PixelGrid = new GType(_PixelGrid = {
         {
             name: "puzzle_changed",
             parameters: [GObject.TYPE_OBJECT]
+        },
+        {
+            name: "highlight_changed",
+            parameters: [GObject.TYPE_OBJECT]
+        },
+        {
+            name: "undo_redo_available",
+            parameters: [GObject.TYPE_BOOLEAN, GObject.TYPE_BOOLEAN]
         }
     ],
 
@@ -69,8 +77,13 @@ PixelGrid = new GType(_PixelGrid = {
     __drawingColor: null,
     _unknownPixelsCount: 0,
 
-    _lastPressedButton: null,
-    _lastButtonPressPosition: null,
+    _pixelUnderPointer: null,
+    _dragStartPosition: null,
+    _dragEndPosition: null,
+    _pixelChangesForUndo: null,
+
+    _undoStack: [],
+    _redoStack: [],
 
     class_init: function(klass, prototype) {
         prototype._BACKGROUND_COLOR = _PixelGrid._BACKGROUND_COLOR;
@@ -81,9 +94,12 @@ PixelGrid = new GType(_PixelGrid = {
         prototype.__drawingColor = _PixelGrid.__drawingColor;
         prototype._unknownPixelsCount = _PixelGrid._unknownPixelsCount;
 
-        prototype._lastPressedButton = _PixelGrid._lastPressedButton;
-        prototype._lastButtonPressPosition =
-            _PixelGrid._lastButtonPressPosition;
+        prototype._pixelUnderPointer = _PixelGrid._pixelUnderPointer;
+        prototype._dragStartPosition = _PixelGrid._dragStartPosition;
+        prototype._dragEndPosition = _PixelGrid._dragEndPosition;
+        prototype._pixelChangesForUndo = _PixelGrid._pixelChangesForUndo;
+        prototype._undoStack = _PixelGrid._undoStack;
+        prototype._redoStack = _PixelGrid._redoStack;
 
         prototype.__defineGetter__("_puzzlePixels",
                                    _PixelGrid._getPuzzlePixels);
@@ -106,22 +122,35 @@ PixelGrid = new GType(_PixelGrid = {
         prototype._getPixelFromWidgetCoordinates =
             _PixelGrid._getPixelFromWidgetCoordinates;
         prototype._onButtonPress = _PixelGrid._onButtonPress;
+        prototype._onButtonRelease = _PixelGrid._onButtonRelease;
         prototype._onMouseMove = _PixelGrid._onMouseMove;
+        prototype._updatePixelUnderPointer =
+            _PixelGrid._updatePixelUnderPointer;
         prototype._setPixelsColor = _PixelGrid._setPixelsColor;
         prototype.revealRemaining = _PixelGrid.revealRemaining;
+        prototype._onStateFlagsChanged = _PixelGrid._onStateFlagsChanged;
+        prototype._storeUndoInformation = _PixelGrid._storeUndoInformation;
+        prototype.undo = _PixelGrid.undo;
+        prototype.redo = _PixelGrid.redo;
+        prototype._emitUndoRedoAvailable =
+            _PixelGrid._emitUndoRedoAvailable;
     },
 
     init: function() {
         this.add_events(Gdk.EventMask.BUTTON_PRESS_MASK);
-        this.add_events(Gdk.EventMask.BUTTON1_MOTION_MASK);
-        this.add_events(Gdk.EventMask.BUTTON2_MOTION_MASK);
-        this.add_events(Gdk.EventMask.BUTTON3_MOTION_MASK);
+        this.add_events(Gdk.EventMask.BUTTON_RELEASE_MASK);
+        this.add_events(Gdk.EventMask.POINTER_MOTION_MASK);
+        this.add_events(Gdk.EventMask.POINTER_MOTION_HINT_MASK);
 
-        this.signal.expose_event.connect(Lang.bind(this, this._onDraw));
-        this.signal.button_press_event.connect(Lang.bind(this,
-                                                 this._onButtonPress));
-        this.signal.motion_notify_event.connect(Lang.bind(this,
-                                                 this._onMouseMove));
+        this.signal.draw.connect(Lang.bind(this, this._onDraw));
+        this.signal.button_press_event.connect(
+            Lang.bind(this, this._onButtonPress));
+        this.signal.button_release_event.connect(
+            Lang.bind(this, this._onButtonRelease));
+        this.signal.motion_notify_event.connect(
+            Lang.bind(this, this._onMouseMove));
+        this.signal.state_flags_changed.connect(
+            Lang.bind(this, this._onStateFlagsChanged));
     },
 
     createBlank: function(columns, rows) {
@@ -139,6 +168,10 @@ PixelGrid = new GType(_PixelGrid = {
             }
         }
 
+        this._undoStack = [];
+        this._redoStack = [];
+        this._emitUndoRedoAvailable(null, null);
+
         this.queue_draw();
         this._unknownPixelsCount = this._numRows * this._numColumns;
     },
@@ -177,10 +210,12 @@ PixelGrid = new GType(_PixelGrid = {
         }
     },
 
-    _onDraw: function(drawingArea) {
-        var cairoContext = new Cairo.Context.from_drawable(drawingArea.window);
+    _onDraw: function(widget, cairoContextStruct) {
+        var cairoContext = new Cairo.Context.steal(cairoContextStruct);
+        this._cairoContext = cairoContext; // FIXME: this is only to keep a ref
+                                           // until the next call
 
-        var size = drawingArea.window.get_size();
+        var size = widget.get_allocation().allocation;
         cairoContext.scale(size.width / this._numColumns,
                            size.height / this._numRows);
 
@@ -200,10 +235,42 @@ PixelGrid = new GType(_PixelGrid = {
             }
         }
 
+        if (this._pixelUnderPointer != null) {
+            var highlightEdgeThickness = 10.0 * this._numColumns / size.width;
+
+            var pattern = new Cairo.LinearGradient(
+                this._pixelUnderPointer.x, 0,
+                this._pixelUnderPointer.x + 1, 0);
+            cairoContext.set_source(pattern);
+            pattern.add_color_stop_rgba(0.0, 1.0, 1.0, 1.0, 1.0);
+            pattern.add_color_stop_rgba(highlightEdgeThickness,
+                                        0.5, 0.5, 0.5, 0.0);
+            pattern.add_color_stop_rgba(1.0 - highlightEdgeThickness,
+                                        0.5, 0.5, 0.5, 0.0);
+            pattern.add_color_stop_rgba(1.0, 1.0, 1.0, 1.0, 1.0);
+            cairoContext.rectangle(this._pixelUnderPointer.x, 0,
+                                   1, this._numRows);
+            cairoContext.fill();
+
+            highlightEdgeThickness = 10.0 * this._numRows / size.height;
+
+            pattern = new Cairo.LinearGradient(
+                0, this._pixelUnderPointer.y,
+                0, this._pixelUnderPointer.y + 1);
+            cairoContext.set_source(pattern);
+            pattern.add_color_stop_rgba(0.0, 1.0, 1.0, 1.0, 1.0);
+            pattern.add_color_stop_rgba(highlightEdgeThickness,
+                                        0.5, 0.5, 0.5, 0.0);
+            pattern.add_color_stop_rgba(1.0 - highlightEdgeThickness,
+                                        0.5, 0.5, 0.5, 0.0);
+            pattern.add_color_stop_rgba(1.0, 1.0, 1.0, 1.0, 1.0);
+            cairoContext.rectangle(0, this._pixelUnderPointer.y,
+                                   this._numColumns, 1);
+            cairoContext.fill();
+        }
+
         cairoContext.set_source_rgb(0.0, 0.0, 0.0);
 
-        // TODO: change the scale to use integer boundaries for line
-        // drawing
         for (i = 0; i < this._numRows + 1; ++i) {
             cairoContext.line_width = 1.0;
             if (i % 5 == 0)
@@ -240,7 +307,7 @@ PixelGrid = new GType(_PixelGrid = {
     },
 
     _getPixelFromWidgetCoordinates: function(x, y) {
-        var widgetSize = this.window.get_size();
+        var widgetSize = this.get_allocation().allocation;
 
         var row = Math.floor(y * this._numRows / widgetSize.height);
         var column = Math.floor(x * this._numColumns / widgetSize.width);
@@ -261,8 +328,10 @@ PixelGrid = new GType(_PixelGrid = {
         if (pixel == null)
             return false;
 
-        this._lastPressedButton = event.button;
-        this._lastButtonPressPosition = pixel;
+        this._dragStartPosition = { x: pixel.x, y: pixel.y };
+        this._dragEndPosition = { x: pixel.x, y: pixel.y };
+        this._pixelChangesForUndo = new PixelChanges();
+        this._pixelChangesForUndo.changeArray = [];
 
         var newColor = this._drawingColor;
         if (event.button == 1)
@@ -274,30 +343,58 @@ PixelGrid = new GType(_PixelGrid = {
         else
             return false;
 
+        var prevColor = this._puzzlePixels[pixel.y][pixel.x];
+        if (prevColor == newColor)
+            return false;
+
         var pixelChanges = new PixelChanges();
-        pixelChanges.changeArray = [{ x: pixel.x,
-                                      y: pixel.y,
-                                      newColor: newColor}];
+        pixelChanges.changeArray = [];
+        var change = { x: pixel.x,
+                       y: pixel.y,
+                       newColor: newColor,
+                       prevColor: prevColor
+                     };
+        pixelChanges.changeArray.push(change);
+        this._pixelChangesForUndo.changeArray.push(change);
 
         this._setPixelsColor(pixelChanges);
 
         return false;
     },
 
+    _onButtonRelease: function(widget, event) {
+        this._dragStartPosition = null;
+        this._dragEndPosition = null;
+
+        if (this._pixelChangesForUndo !== null &&
+            this._pixelChangesForUndo.changeArray.length != 0) {
+            this._storeUndoInformation(this._pixelChangesForUndo);
+        }
+
+        this._pixelChangesForUndo = null;
+
+        return false;
+    },
+
     _onMouseMove: function(widget, event) {
         event = event.motion;
 
-        var newColor = this._drawingColor;
-        if (this._lastPressedButton == 1)
-            newColor = this._drawingColor;
-        else if (this._lastPressedButton == 3)
-            newColor = this._BACKGROUND_COLOR;
-        else if (this._lastPressedButton == 2)
-            newColor = this._NO_COLOR;
-        else
-            return false;
+        this._updatePixelUnderPointer(event);
 
-        var widgetSize = this.window.get_size();
+        var newColor = this._drawingColor;
+        if (event.state & Gdk.ModifierType.BUTTON1_MASK)
+            newColor = this._drawingColor;
+        else if (event.state & Gdk.ModifierType.BUTTON3_MASK)
+            newColor = this._BACKGROUND_COLOR;
+        else if (event.state & Gdk.ModifierType.BUTTON2_MASK)
+            newColor = this._NO_COLOR;
+        else {
+            if (event.is_hint == true)
+                Gdk.event_request_motions(event);
+            return false;
+        }
+
+        var widgetSize = this.get_allocation().allocation;
         if (event.x < 0)
             event.x = 0;
 
@@ -311,50 +408,96 @@ PixelGrid = new GType(_PixelGrid = {
             event.y = widgetSize.height - 1;
 
         var pixel = this._getPixelFromWidgetCoordinates(event.x, event.y);
-        if (pixel == null)
+        if (pixel == null) {
+            if (event.is_hint == true)
+                Gdk.event_request_motions(event);
             return false;
+        }
 
-        var xDelta = pixel.x - this._lastButtonPressPosition.x;
-        var yDelta = pixel.y - this._lastButtonPressPosition.y;
+        var xDelta = pixel.x - this._dragStartPosition.x;
+        var yDelta = pixel.y - this._dragStartPosition.y;
+
+        if (xDelta == 0 && yDelta == 0) {
+            if (event.is_hint == true)
+                Gdk.event_request_motions(event);
+            return false;
+        }
+
+        var extendDragBoundaries = function(isVertical) {
+            var direction = "x";
+            if (isVertical)
+                direction = "y";
+
+            this._dragEndPosition[direction] =
+                Math.max(pixel[direction], this._dragEndPosition[direction]);
+            this._dragStartPosition[direction] =
+                Math.min(pixel[direction], this._dragStartPosition[direction]);
+        };
+
+        var isVertical;
+        if (this._dragStartPosition.x == this._dragEndPosition.x &&
+            this._dragStartPosition.y == this._dragEndPosition.y) {
+            // First time, set the orientation of the drag based on
+            // magnitudes of delta movement in both directions
+            isVertical = Math.abs(yDelta) > Math.abs(xDelta);
+        } else {
+            // Once set, use the direction and only extend it in that
+            // direction
+            isVertical = (this._dragStartPosition.x == this._dragEndPosition.x);
+        }
+        extendDragBoundaries.call(this, isVertical);
 
         var pixelChanges = new PixelChanges();
         pixelChanges.changeArray = [];
 
-        if (Math.abs(yDelta) > Math.abs(xDelta)) {
-            if (yDelta == 0)
-                return false;
+        var increment = Math.round(yDelta / Math.abs(yDelta));
 
-            var increment = Math.round(yDelta / Math.abs(yDelta));
-            for (var i = this._lastButtonPressPosition.y;
-                 i != (pixel.y + increment);
-                 i += increment) {
-                pixelChanges.changeArray.push({
-                    x: this._lastButtonPressPosition.x,
-                    y: i,
-                    newColor: newColor
-                });
-            }
-        } else {
-            if (xDelta == 0)
-                return false;
+        for (var y = this._dragStartPosition.y;
+             y <= this._dragEndPosition.y;
+             ++y) {
+            for (var x = this._dragStartPosition.x;
+                 x <= this._dragEndPosition.x;
+                 ++x){
+                if (this._puzzlePixels[y][x] == newColor)
+                    continue;
 
-            var increment = Math.round(xDelta / Math.abs(xDelta));
-            for (var i = this._lastButtonPressPosition.x;
-                 i != (pixel.x + increment);
-                 i += increment) {
-                pixelChanges.changeArray.push({
-                    x: i,
-                    y: this._lastButtonPressPosition.y,
-                    newColor: newColor
-                });
+                var change = {
+                    x: x,
+                    y: y,
+                    newColor: newColor,
+                    prevColor: this._puzzlePixels[y][x]
+                };
+                pixelChanges.changeArray.push(change);
+                this._pixelChangesForUndo.changeArray.push(change);
             }
         }
 
         this._setPixelsColor(pixelChanges);
 
+        if (event.is_hint == true)
+            Gdk.event_request_motions(event);
+
         return false;
     },
 
+    _updatePixelUnderPointer: function(event) {
+        var widgetSize = this.get_allocation().allocation;
+        var pixelX = Math.floor(event.x * this._numColumns / widgetSize.width);
+        var pixelY = Math.floor(event.y * this._numRows / widgetSize.height);
+        if (this._pixelUnderPointer != null &&
+            this._pixelUnderPointer.x == pixelX &&
+            this._pixelUnderPointer.y == pixelY) {
+            return;
+        }
+
+        this._pixelUnderPointer = new Pixel();
+        this._pixelUnderPointer.x = pixelX;
+        this._pixelUnderPointer.y = pixelY;
+        this.signal.highlight_changed.emit(this._pixelUnderPointer);
+
+        this.queue_draw();
+    },
+
     _setPixelsColor: function(pixelChanges) {
         var emitPixelChanges = new PixelChanges();
         emitPixelChanges.puzzlePixels = this._puzzlePixels;
@@ -362,8 +505,6 @@ PixelGrid = new GType(_PixelGrid = {
 
         for (var i = 0; i < pixelChanges.changeArray.length; ++i) {
             var change = pixelChanges.changeArray[i];
-            change.prevColor = this._puzzlePixels[change.y][change.x];
-
             if (change.prevColor == change.newColor)
                 continue;
 
@@ -397,6 +538,93 @@ PixelGrid = new GType(_PixelGrid = {
         }
 
         this.queue_draw();
+    },
+
+    _onStateFlagsChanged: function(widget, previousState) {
+        var previousIsSensitive = !(previousState &
+                                    Gtk.StateFlags.INSENSITIVE);
+        var currentIsSensitive = this.sensitive;
+        if (previousIsSensitive == true && currentIsSensitive == false) {
+            this._pixelUnderPointer = null;
+            this.signal.highlight_changed.emit(this._pixelUnderPointer);
+            this.queue_draw();
+        }
+    },
+
+    _storeUndoInformation: function(pixelChanges) {
+        var previousUndoStackLength = this._undoStack.length;
+        var previousRedoStackLength = this._redoStack.length;
+
+        this._redoStack = [];
+        this._undoStack.push(pixelChanges);
+
+        this._emitUndoRedoAvailable(previousUndoStackLength,
+                                    previousRedoStackLength);
+    },
+
+    undo: function() {
+        if (this.sensitive == false)
+            return;
+
+        if (this._undoStack.length == 0)
+            return;
+
+        var previousUndoStackLength = this._undoStack.length;
+        var previousRedoStackLength = this._redoStack.length;
+
+        var pixelChanges = this._undoStack.pop();
+        this._redoStack.push(pixelChanges);
+
+        var undoPixelChanges = new PixelChanges();
+        undoPixelChanges.changeArray = [];
+        pixelChanges.changeArray.forEach(
+            function(change) {
+                var newChange = {};
+                for (var key in change) {
+                    newChange[key] = change[key];
+                }
+
+                newChange.prevColor = change.newColor;
+                newChange.newColor = change.prevColor;
+                undoPixelChanges.changeArray.push(newChange);
+            }
+        );
+
+        this._setPixelsColor(undoPixelChanges);
+
+        this._emitUndoRedoAvailable(previousUndoStackLength,
+                                    previousRedoStackLength);
+    },
+
+    redo: function() {
+        if (this.sensitive == false)
+            return;
+
+        if (this._redoStack.length == 0)
+            return;
+
+        var previousUndoStackLength = this._undoStack.length;
+        var previousRedoStackLength = this._redoStack.length;
+
+        var pixelChanges = this._redoStack.pop();
+        this._undoStack.push(pixelChanges);
+        this._setPixelsColor(pixelChanges);
+
+        this._emitUndoRedoAvailable(previousUndoStackLength,
+                                    previousRedoStackLength);
+    },
+
+    _emitUndoRedoAvailable: function(previousUndoStackLength,
+                                     previousRedoStackLength) {
+        if ((this._undoStack.length == 0 && previousUndoStackLength != 0) ||
+            (this._undoStack.length != 0 && previousUndoStackLength == 0) ||
+            (this._redoStack.length == 0 && previousRedoStackLength != 0) ||
+            (this._redoStack.length != 0 && previousRedoStackLength == 0)) {
+            var undoAvailable = (this._undoStack.length != 0);
+            var redoAvailable = (this._redoStack.length != 0);
+            this.signal.undo_redo_available.emit(undoAvailable,
+                                                 redoAvailable);
+        }
     }
 });
 
@@ -404,3 +632,8 @@ PixelChanges = new GType(_PixelChanges = {
     parent: GObject.TYPE_OBJECT,
     name: "PixelChanges"
 });
+
+Pixel = new GType(_Pixel = {
+    parent: GObject.TYPE_OBJECT,
+    name: "Pixel"
+});
